@@ -2,7 +2,7 @@ package Mojolicious::Plugin::TagHelpers;
 use Mojo::Base 'Mojolicious::Plugin';
 
 use Mojo::ByteStream;
-use Mojo::Util qw(deprecated xml_escape);
+use Mojo::Util 'xml_escape';
 use Scalar::Util 'blessed';
 
 sub register {
@@ -44,17 +44,17 @@ sub register {
 }
 
 sub _csrf_field {
-  my $self = shift;
-  return $self->hidden_field(csrf_token => $self->csrf_token, @_);
+  my $c = shift;
+  return _hidden_field($c, csrf_token => $c->helpers->csrf_token, @_);
 }
 
 sub _form_for {
-  my ($self, @url) = (shift, shift);
+  my ($c, @url) = (shift, shift);
   push @url, shift if ref $_[0] eq 'HASH';
 
   # POST detection
   my @post;
-  if (my $r = $self->app->routes->lookup($url[0])) {
+  if (my $r = $c->app->routes->lookup($url[0])) {
     my %methods = (GET => 1, POST => 1);
     do {
       my @via = @{$r->via || []};
@@ -63,21 +63,20 @@ sub _form_for {
     @post = (method => 'POST') if $methods{POST} && !$methods{GET};
   }
 
-  return _tag('form', action => $self->url_for(@url), @post, @_);
+  return _tag('form', action => $c->url_for(@url), @post, @_);
 }
 
 sub _hidden_field {
-  my $self = shift;
-  my %attrs = (name => shift, value => shift, @_, type => 'hidden');
-  return _tag('input', %attrs);
+  my $c = shift;
+  return _tag('input', name => shift, value => shift, @_, type => 'hidden');
 }
 
 sub _input {
-  my ($self, $name) = (shift, shift);
+  my ($c, $name) = (shift, shift);
   my %attrs = @_ % 2 ? (value => shift, @_) : @_;
 
   # Special selection value
-  my @values = $self->param($name);
+  my @values = @{$c->every_param($name)};
   my $type = $attrs{type} || '';
   if (@values && $type ne 'submit') {
 
@@ -92,11 +91,11 @@ sub _input {
     else { $attrs{value} = $values[0] }
   }
 
-  return _validation($self, $name, 'input', %attrs, name => $name);
+  return _validation($c, $name, 'input', %attrs, name => $name);
 }
 
 sub _javascript {
-  my $self = shift;
+  my $c = shift;
 
   # CDATA
   my $cb = sub {''};
@@ -105,19 +104,19 @@ sub _javascript {
   }
 
   # URL
-  my $src = @_ % 2 ? $self->url_for(shift) : undef;
+  my $src = @_ % 2 ? $c->url_for(shift) : undef;
 
   return _tag('script', @_, $src ? (src => $src) : (), $cb);
 }
 
 sub _label_for {
-  my ($self, $name) = (shift, shift);
+  my ($c, $name) = (shift, shift);
   my $content = ref $_[-1] eq 'CODE' ? pop : shift;
-  return _validation($self, $name, 'label', for => $name, @_, $content);
+  return _validation($c, $name, 'label', for => $name, @_, $content);
 }
 
 sub _link_to {
-  my ($self, $content) = (shift, shift);
+  my ($c, $content) = (shift, shift);
   my @url = ($content);
 
   # Content
@@ -129,7 +128,7 @@ sub _link_to {
   # Captures
   push @url, shift if ref $_[0] eq 'HASH';
 
-  return _tag('a', href => $self->url_for(@url), @_);
+  return _tag('a', href => $c->url_for(@url), @_);
 }
 
 sub _option {
@@ -145,43 +144,36 @@ sub _option {
 }
 
 sub _password_field {
-  my ($self, $name) = (shift, shift);
-  return _validation($self, $name, 'input', @_, name => $name,
+  my ($c, $name) = (shift, shift);
+  return _validation($c, $name, 'input', @_, name => $name,
     type => 'password');
 }
 
 sub _select_field {
-  my ($self, $name, $options, %attrs) = (shift, shift, shift, @_);
+  my ($c, $name, $options, %attrs) = (shift, shift, shift, @_);
 
-  my %values = map { $_ => 1 } $self->param($name);
+  my %values = map { $_ => 1 } @{$c->every_param($name)};
 
   my $groups = '';
   for my $group (@$options) {
 
-    # DEPRECATED in Top Hat!
-    if (ref $group eq 'HASH') {
-      deprecated
-        'hash references are DEPRECATED in favor of Mojo::Collection objects';
-      $group = Mojo::Collection->new(each %$group);
-    }
-
     # "optgroup" tag
     if (blessed $group && $group->isa('Mojo::Collection')) {
-      my ($label, $values) = splice @$group, 0, 2;
+      my ($label, $values, %attrs) = @$group;
       my $content = join '', map { _option(\%values, $_) } @$values;
-      $groups .= _tag('optgroup', label => $label, @$group, sub {$content});
+      $groups .= _tag('optgroup', label => $label, %attrs, sub {$content});
     }
 
     # "option" tag
     else { $groups .= _option(\%values, $group) }
   }
 
-  return _validation($self, $name, 'select', %attrs, name => $name,
+  return _validation($c, $name, 'select', %attrs, name => $name,
     sub {$groups});
 }
 
 sub _stylesheet {
-  my $self = shift;
+  my $c = shift;
 
   # CDATA
   my $cb;
@@ -190,14 +182,14 @@ sub _stylesheet {
   }
 
   # "link" or "style" tag
-  my $href = @_ % 2 ? $self->url_for(shift) : undef;
+  my $href = @_ % 2 ? $c->url_for(shift) : undef;
   return $href
     ? _tag('link', rel => 'stylesheet', href => $href, @_)
     : _tag('style', @_, $cb);
 }
 
 sub _submit_button {
-  my $self = shift;
+  my $c = shift;
   return _tag('input', value => shift // 'Ok', @_, type => 'submit');
 }
 
@@ -213,6 +205,13 @@ sub _tag {
 
   # Attributes
   my %attrs = @_;
+  if ($attrs{data} && ref $attrs{data} eq 'HASH') {
+    while (my ($key, $value) = each %{$attrs{data}}) {
+      $key =~ y/_/-/;
+      $attrs{lc("data-$key")} = $value;
+    }
+    delete $attrs{data};
+  }
   $tag .= qq{ $_="} . xml_escape($attrs{$_} // '') . '"' for sort keys %attrs;
 
   # Empty element
@@ -226,28 +225,28 @@ sub _tag {
 }
 
 sub _tag_with_error {
-  my ($self, $tag) = (shift, shift);
+  my ($c, $tag) = (shift, shift);
   my ($content, %attrs) = (@_ % 2 ? pop : undef, @_);
   $attrs{class} .= $attrs{class} ? ' field-with-error' : 'field-with-error';
   return _tag($tag, %attrs, defined $content ? $content : ());
 }
 
 sub _text_area {
-  my ($self, $name) = (shift, shift);
+  my ($c, $name) = (shift, shift);
 
   # Make sure content is wrapped
   my $cb = ref $_[-1] eq 'CODE' ? pop : sub {''};
   my $content = @_ % 2 ? shift : undef;
   $cb = sub { xml_escape $content }
-    if defined($content = $self->param($name) // $content);
+    if defined($content = $c->param($name) // $content);
 
-  return _validation($self, $name, 'textarea', @_, name => $name, $cb);
+  return _validation($c, $name, 'textarea', @_, name => $name, $cb);
 }
 
 sub _validation {
-  my ($self, $name) = (shift, shift);
-  return _tag(@_) unless $self->validation->has_error($name);
-  return $self->tag_with_error(@_);
+  my ($c, $name) = (shift, shift);
+  return _tag(@_) unless $c->validation->has_error($name);
+  return $c->helpers->tag_with_error(@_);
 }
 
 1;
@@ -290,6 +289,9 @@ C<tag_with_error> helper, to make styling with CSS easier.
 This is a core plugin, that means it is always enabled and its code a good
 example for learning how to build new plugins, you're welcome to fork it.
 
+See L<Mojolicious::Plugins/"PLUGINS"> for a list of plugins that are available
+by default.
+
 =head1 HELPERS
 
 L<Mojolicious::Plugin::TagHelpers> implements the following helpers.
@@ -297,13 +299,13 @@ L<Mojolicious::Plugin::TagHelpers> implements the following helpers.
 =head2 check_box
 
   %= check_box employed => 1
-  %= check_box employed => 1, id => 'foo'
+  %= check_box employed => 1, disabled => 'disabled'
 
-Generate checkbox input element. Previous input values will automatically get
-picked up and shown as default.
+Generate C<input> tag of type C<checkbox>. Previous input values will
+automatically get picked up and shown as default.
 
   <input name="employed" type="checkbox" value="1" />
-  <input id="foo" name="employed" type="checkbox" value="1" />
+  <input disabled="disabled" name="employed" type="checkbox" value="1" />
 
 =head2 color_field
 
@@ -311,8 +313,8 @@ picked up and shown as default.
   %= color_field background => '#ffffff'
   %= color_field background => '#ffffff', id => 'foo'
 
-Generate color input element. Previous input values will automatically get
-picked up and shown as default.
+Generate C<input> tag of type C<color>. Previous input values will
+automatically get picked up and shown as default.
 
   <input name="background" type="color" />
   <input name="background" type="color" value="#ffffff" />
@@ -322,7 +324,7 @@ picked up and shown as default.
 
   %= csrf_field
 
-Generate hidden input element with
+Generate C<input> tag of type C<hidden> with
 L<Mojolicious::Plugin::DefaultHelpers/"csrf_token">.
 
   <input name="csrf_token" type="hidden" value="fa6a08..." />
@@ -333,8 +335,8 @@ L<Mojolicious::Plugin::DefaultHelpers/"csrf_token">.
   %= date_field end => '2012-12-21'
   %= date_field end => '2012-12-21', id => 'foo'
 
-Generate date input element. Previous input values will automatically get
-picked up and shown as default.
+Generate C<input> tag of type C<date>. Previous input values will
+automatically get picked up and shown as default.
 
   <input name="end" type="date" />
   <input name="end" type="date" value="2012-12-21" />
@@ -346,8 +348,8 @@ picked up and shown as default.
   %= datetime_field end => '2012-12-21T23:59:59Z'
   %= datetime_field end => '2012-12-21T23:59:59Z', id => 'foo'
 
-Generate datetime input element. Previous input values will automatically get
-picked up and shown as default.
+Generate C<input> tag of type C<datetime>. Previous input values will
+automatically get picked up and shown as default.
 
   <input name="end" type="datetime" />
   <input name="end" type="datetime" value="2012-12-21T23:59:59Z" />
@@ -359,8 +361,8 @@ picked up and shown as default.
   %= email_field notify => 'nospam@example.com'
   %= email_field notify => 'nospam@example.com', id => 'foo'
 
-Generate email input element. Previous input values will automatically get
-picked up and shown as default.
+Generate C<input> tag of type C<email>. Previous input values will
+automatically get picked up and shown as default.
 
   <input name="notify" type="email" />
   <input name="notify" type="email" value="nospam@example.com" />
@@ -371,7 +373,7 @@ picked up and shown as default.
   %= file_field 'avatar'
   %= file_field 'avatar', id => 'foo'
 
-Generate file input element.
+Generate C<input> tag of type C<file>.
 
   <input name="avatar" type="file" />
   <input id="foo" name="avatar" type="file" />
@@ -386,8 +388,8 @@ Generate file input element.
     %= text_field 'first_name'
     %= submit_button
   % end
-  %= form_for '/login' => (method => 'POST') => begin
-    %= text_field 'first_name'
+  %= form_for '/login' => (enctype => 'multipart/form-data') => begin
+    %= text_field 'first_name', disabled => 'disabled'
     %= submit_button
   % end
   %= form_for 'http://example.com/login' => (method => 'POST') => begin
@@ -395,24 +397,24 @@ Generate file input element.
     %= submit_button
   % end
 
-Generate portable form tag with L<Mojolicious::Controller/"url_for">. For
-routes that allow POST but not GET, a C<method> attribute will be
+Generate portable C<form> tag with L<Mojolicious::Controller/"url_for">. For
+routes that allow C<POST> but not C<GET>, a C<method> attribute will be
 automatically added.
 
   <form action="/path/to/login">
-    <input name="first_name" />
+    <input name="first_name" type="text" />
     <input value="Ok" type="submit" />
   </form>
   <form action="/path/to/login.txt" method="POST">
-    <input name="first_name" />
+    <input name="first_name" type="text" />
     <input value="Ok" type="submit" />
   </form>
-  <form action="/path/to/login" method="POST">
-    <input name="first_name" />
+  <form action="/path/to/login" enctype="multipart/form-data">
+    <input disabled="disabled" name="first_name" type="text" />
     <input value="Ok" type="submit" />
   </form>
   <form action="http://example.com/login" method="POST">
-    <input name="first_name" />
+    <input name="first_name" type="text" />
     <input value="Ok" type="submit" />
   </form>
 
@@ -421,7 +423,7 @@ automatically added.
   %= hidden_field foo => 'bar'
   %= hidden_field foo => 'bar', id => 'bar'
 
-Generate hidden input element.
+Generate C<input> tag of type C<hidden>.
 
   <input name="foo" type="hidden" value="bar" />
   <input id="bar" name="foo" type="hidden" value="bar" />
@@ -431,7 +433,7 @@ Generate hidden input element.
   %= image '/images/foo.png'
   %= image '/images/foo.png', alt => 'Foo'
 
-Generate portable img tag.
+Generate portable C<img> tag.
 
   <img src="/path/to/images/foo.png" />
   <img alt="Foo" src="/path/to/images/foo.png" />
@@ -442,8 +444,8 @@ Generate portable img tag.
   %= input_tag first_name => 'Default name'
   %= input_tag 'employed', type => 'checkbox'
 
-Generate form input element. Previous input values will automatically get
-picked up and shown as default.
+Generate C<input> tag. Previous input values will automatically get picked up
+and shown as default.
 
   <input name="first_name" />
   <input name="first_name" value="Default name" />
@@ -456,7 +458,7 @@ picked up and shown as default.
     var a = 'b';
   % end
 
-Generate portable script tag for C<Javascript> asset.
+Generate portable C<script> tag for JavaScript asset.
 
   <script src="/path/to/script.js" />
   <script><![CDATA[
@@ -474,7 +476,7 @@ Generate portable script tag for C<Javascript> asset.
     First name
   % end
 
-Generate label.
+Generate C<label> tag.
 
   <label for="first_name">First name</label>
   <label class="user" for="first_name">First name</label>
@@ -498,8 +500,8 @@ Generate label.
   <%= link_to 'http://mojolicio.us' => begin %>Mojolicious<% end %>
   <%= link_to url_for->query(foo => 'bar')->to_abs => begin %>Retry<% end %>
 
-Generate portable link with L<Mojolicious::Controller/"url_for">, defaults to
-using the capitalized link target as content.
+Generate portable C<a> tag with L<Mojolicious::Controller/"url_for">, defaults
+to using the capitalized link target as content.
 
   <a href="/path/to/index">Home</a>
   <a class="menu" href="/path/to/index.txt">Home</a>
@@ -518,8 +520,8 @@ using the capitalized link target as content.
   %= month_field vacation => '2012-12'
   %= month_field vacation => '2012-12', id => 'foo'
 
-Generate month input element. Previous input values will automatically get
-picked up and shown as default.
+Generate C<input> tag of type C<month>. Previous input values will
+automatically get picked up and shown as default.
 
   <input name="vacation" type="month" />
   <input name="vacation" type="month" value="2012-12" />
@@ -531,8 +533,8 @@ picked up and shown as default.
   %= number_field age => 25
   %= number_field age => 25, id => 'foo', min => 0, max => 200
 
-Generate number input element. Previous input values will automatically get
-picked up and shown as default.
+Generate C<input> tag of type C<number>. Previous input values will
+automatically get picked up and shown as default.
 
   <input name="age" type="number" />
   <input name="age" type="number" value="25" />
@@ -543,7 +545,7 @@ picked up and shown as default.
   %= password_field 'pass'
   %= password_field 'pass', id => 'foo'
 
-Generate password input element.
+Generate C<input> tag of type C<password>.
 
   <input name="pass" type="password" />
   <input id="foo" name="pass" type="password" />
@@ -553,8 +555,8 @@ Generate password input element.
   %= radio_button country => 'germany'
   %= radio_button country => 'germany', id => 'foo'
 
-Generate radio input element. Previous input values will automatically get
-picked up and shown as default.
+Generate C<input> tag of type C<radio>. Previous input values will
+automatically get picked up and shown as default.
 
   <input name="country" type="radio" value="germany" />
   <input id="foo" name="country" type="radio" value="germany" />
@@ -565,8 +567,8 @@ picked up and shown as default.
   %= range_field age => 25
   %= range_field age => 25, id => 'foo', min => 0, max => 200
 
-Generate range input element. Previous input values will automatically get
-picked up and shown as default.
+Generate C<input> tag of type C<range>. Previous input values will
+automatically get picked up and shown as default.
 
   <input name="age" type="range" />
   <input name="age" type="range" value="25" />
@@ -578,8 +580,8 @@ picked up and shown as default.
   %= search_field q => 'perl'
   %= search_field q => 'perl', id => 'foo'
 
-Generate search input element. Previous input values will automatically get
-picked up and shown as default.
+Generate C<input> tag of type C<search>. Previous input values will
+automatically get picked up and shown as default.
 
   <input name="q" type="search" />
   <input name="q" type="search" value="perl" />
@@ -589,12 +591,12 @@ picked up and shown as default.
 
   %= select_field country => [qw(de en)]
   %= select_field country => [[Germany => 'de'], 'en'], id => 'eu'
-  %= select_field country => [[Germany => 'de', class => 'eu'], 'en']
+  %= select_field country => [[Germany => 'de', disabled => 'disabled'], 'en']
   %= select_field country => [c(EU => [[Germany => 'de'], 'en'], id => 'eu')]
   %= select_field country => [c(EU => [qw(de en)]), c(Asia => [qw(cn jp)])]
 
-Generate select and option elements from array references and optgroup
-elements from L<Mojo::Collection> objects. Previous input values will
+Generate C<select> and C<option> tags from array references and C<optgroup>
+tags from L<Mojo::Collection> objects. Previous input values will
 automatically get picked up and shown as default.
 
   <select name="country">
@@ -606,7 +608,7 @@ automatically get picked up and shown as default.
     <option value="en">en</option>
   </select>
   <select name="country">
-    <option class="eu" value="de">Germany</option>
+    <option disabled="disabled" value="de">Germany</option>
     <option value="en">en</option>
   </select>
   <select name="country">
@@ -633,7 +635,7 @@ automatically get picked up and shown as default.
     body {color: #000}
   % end
 
-Generate portable style or link tag for C<CSS> asset.
+Generate portable C<style> or C<link> tag for CSS asset.
 
   <link href="/path/to/foo.css" rel="stylesheet" />
   <style><![CDATA[
@@ -645,49 +647,52 @@ Generate portable style or link tag for C<CSS> asset.
   %= submit_button
   %= submit_button 'Ok!', id => 'foo'
 
-Generate submit input element.
+Generate C<input> tag of type C<submit>.
 
   <input type="submit" value="Ok" />
   <input id="foo" type="submit" value="Ok!" />
 
 =head2 t
 
-  %=t div => 'some & content'
+  %=t div => 'test & 123'
 
 Alias for L</"tag">.
 
-  <div>some &amp; content</div>
+  <div>test &amp; 123</div>
 
 =head2 tag
 
   %= tag 'div'
   %= tag 'div', id => 'foo'
-  %= tag div => 'some & content'
-  %= tag div => (id => 'foo') => 'some & content'
+  %= tag div => 'test & 123'
+  %= tag div => (id => 'foo') => 'test & 123'
+  %= tag div => (data => {my_id => 1, Name => 'test'}) => 'test & 123'
   %= tag div => begin
-    some & content
+    test & 123
   % end
-  <%= tag div => (id => 'foo') => begin %>some & content<% end %>
+  <%= tag div => (id => 'foo') => begin %>test & 123<% end %>
 
 HTML/XML tag generator.
 
   <div />
   <div id="foo" />
-  <div>some &amp; content</div>
-  <div id="foo">some &amp; content</div>
+  <div>test &amp; 123</div>
+  <div id="foo">test &amp; 123</div>
+  <div data-my-id="1" data-name="test">test &amp; 123</div>
   <div>
-    some & content
+    test & 123
   </div>
-  <div id="foo">some & content</div>
+  <div id="foo">test & 123</div>
 
 Very useful for reuse in more specific tag helpers.
 
-  $self->tag('div');
-  $self->tag('div', id => 'foo');
-  $self->tag(div => sub { 'Content' });
+  my $output = $c->tag('div');
+  my $output = $c->tag('div', id => 'foo');
+  my $output = $c->tag(div => '<p>This will be escaped</p>');
+  my $output = $c->tag(div => sub { '<p>This will not be escaped</p>' });
 
 Results are automatically wrapped in L<Mojo::ByteStream> objects to prevent
-accidental double escaping.
+accidental double escaping in C<ep> templates.
 
 =head2 tag_with_error
 
@@ -703,8 +708,8 @@ Same as L</"tag">, but adds the class C<field-with-error>.
   %= tel_field work => '123456789'
   %= tel_field work => '123456789', id => 'foo'
 
-Generate tel input element. Previous input values will automatically get
-picked up and shown as default.
+Generate C<input> tag of type C<tel>. Previous input values will automatically
+get picked up and shown as default.
 
   <input name="work" type="tel" />
   <input name="work" type="tel" value="123456789" />
@@ -719,7 +724,7 @@ picked up and shown as default.
     Default!
   % end
 
-Generate textarea element. Previous input values will automatically get picked
+Generate C<textarea> tag. Previous input values will automatically get picked
 up and shown as default.
 
   <textarea name="foo"></textarea>
@@ -735,8 +740,8 @@ up and shown as default.
   %= text_field first_name => 'Default name'
   %= text_field first_name => 'Default name', class => 'user'
 
-Generate text input element. Previous input values will automatically get
-picked up and shown as default.
+Generate C<input> tag of type C<text>. Previous input values will
+automatically get picked up and shown as default.
 
   <input name="first_name" type="text" />
   <input name="first_name" type="text" value="Default name" />
@@ -748,8 +753,8 @@ picked up and shown as default.
   %= time_field start => '23:59:59'
   %= time_field start => '23:59:59', id => 'foo'
 
-Generate time input element. Previous input values will automatically get
-picked up and shown as default.
+Generate C<input> tag of type C<time>. Previous input values will
+automatically get picked up and shown as default.
 
   <input name="start" type="time" />
   <input name="start" type="time" value="23:59:59" />
@@ -761,8 +766,8 @@ picked up and shown as default.
   %= url_field address => 'http://mojolicio.us'
   %= url_field address => 'http://mojolicio.us', id => 'foo'
 
-Generate url input element. Previous input values will automatically get
-picked up and shown as default.
+Generate C<input> tag of type C<url>. Previous input values will automatically
+get picked up and shown as default.
 
   <input name="address" type="url" />
   <input name="address" type="url" value="http://mojolicio.us" />
@@ -774,8 +779,8 @@ picked up and shown as default.
   %= week_field vacation => '2012-W17'
   %= week_field vacation => '2012-W17', id => 'foo'
 
-Generate week input element. Previous input values will automatically get
-picked up and shown as default.
+Generate C<input> tag of type C<week>. Previous input values will
+automatically get picked up and shown as default.
 
   <input name="vacation" type="week" />
   <input name="vacation" type="week" value="2012-W17" />
